/*
    Copyright (C) 2001-2003 Paul Davis 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: curve.h,v 1.14 2003/08/06 16:05:23 pbd Exp $
*/

#ifndef __ardour_curve_h__
#define __ardour_curve_h__

#include <sys/types.h>
#include <sigc++/signal_system.h>
#include <pbd/lockmonitor.h>
#include <pbd/undo.h>
#include <vector>
#include <pthread.h>

namespace ARDOUR {

class Curve : public SigC::Object
{
  public:
	Curve (double min_yval, double max_yval, float defaultvalue);
	Curve (const Curve&);
	~Curve ();
	Curve& operator= (const Curve&);

	UndoAction get_memento ();
	SigC::Signal0<void> Changed;

	void get_vector (double x0, double x1, float *arg, size_t veclen);
	int add_control_point (double x, double y, bool allow_resize = true);
	void remove_control_point (size_t);
	void remove_control_points (size_t, size_t);
	void reset_control_point (size_t n, double xval, double yval);
	void reset_control_points (size_t, size_t, double);
	void truncate_end (double last_coordinate);
	void truncate_start (double overall_length);

	void finish_resize ();

	void set_max_xval (double);
	double get_max_xval() const { return max_xval; }

	size_t n_control_points() const { return npoints; }
	
	bool closest_control_point_before (double xval, size_t& nth);
	bool closest_control_point_after (double xval, size_t& nth);
	bool control_points_adjacent (double xval, size_t& before, size_t& after,
				      double& before_x, double& after_x);

	bool nth (size_t n, double& xval, double &yval) {
		if (n < npoints) {
			xval = x[n];
			yval = y[n];
			return true;
		}
		return false;
	}

	void clear ();
	void solve ();

	template<class T> void apply_to_points (T& obj, void (T::*method)(size_t, double *, double *y)) {
		(obj.*method)(npoints, x, y);
		mark_dirty ();
	}

	void set_clear_on_write (bool yn);

  private:
	double *x;
	double *y;
	double *coeff;
	size_t npoints;
	double min_x;
	double max_x;
	double min_yval;
	double max_yval;
	double max_xval;
	float  default_value;
	size_t size;
	PBD::Lock lock;
	bool _dirty;
	size_t last_lo;
	size_t last_hi;
	double last_added_point_x;
	bool   clear_on_write;
	bool   resize_pending;

	inline void spline_bound (double val, size_t& lo, size_t& hi) {
		
		if (val >= x[last_lo] && val <= x[last_hi]) {
		    lo = last_lo;
		    hi = last_hi;
		    return;
		}

		/* do a binary search for the right interval */

		size_t k_lo = 0; 
		size_t k_hi = npoints - 1;
		size_t k;

		while (k_hi - k_lo > 1) {

			k = (k_hi + k_lo) / 2;

			if (x[k] > val) {
				k_hi = k;
			} else {
				k_lo = k;
			}
		}

		lo = k_lo;
		hi = k_hi;
		last_lo = lo;
		last_hi = hi;
	}

	inline double spline_eval (double val)
	{
		size_t lo, hi;
		double val2;
		spline_bound (val, lo, hi);
		
		val2 = val * val;
		return coeff[hi*4] + (coeff[hi*4+1] * val) + (coeff[hi*4+2] * val2) + (coeff[hi*4+3] * val2 * val);
	}

	void mark_dirty ();
	void copy_from_pointer (const Curve* copy);
};

}; /* namespace ARDOUR */

#endif /* __ardour_curve_h__ */
